use std::{path::PathBuf, time::Duration};

const AVERAGE: &str = "Average";
const TOTAL: &str = "total";

use crate::{
    benchmarking::ConductedBenchmark,
    cli::{GivenDay, GivenTask},
    constants,
    solving_given::NoSolutionFound,
};

pub const COLUMN_NUMBER: usize = 6;

#[derive(Debug, Default)]
pub struct RowBuilder {
    day: Option<String>,
    task: Option<String>,
    taken_time: Option<Duration>,
    actual_result: Option<String>,
    expected_result: Option<String>,
    path_to_input: Option<String>,
}

impl From<NoSolutionFound> for RowBuilder {
    fn from(value: NoSolutionFound) -> Self {
        match value {
            NoSolutionFound::DayNotFound(day) => Self::no_day_found(day),
            NoSolutionFound::TaskNotFound { day, task } => Self::no_task_found(day, task),
        }
    }
}

impl From<ConductedBenchmark> for RowBuilder {
    fn from(value: ConductedBenchmark) -> Self {
        Self::new(
            value.day(),
            value.task(),
            value.how_long(),
            value
                .maybe_different_actual_output()
                .map(ToString::to_string)
                .unwrap_or_default(),
            value.expected_output().to_string(),
            value.path_of_input().to_path_buf(),
        )
    }
}

impl RowBuilder {
    pub fn new(
        day: GivenDay,
        task: GivenTask,
        taken_time: Duration,
        actual_result: String,
        expected_result: String,
        path: PathBuf,
    ) -> Self {
        Self {
            day: Some(day.to_string()),
            task: Some(task.to_string()),
            taken_time: Some(taken_time),
            actual_result: Some(actual_result),
            expected_result: Some(expected_result),
            path_to_input: Some(path.to_string_lossy().to_string()),
        }
    }

    pub fn no_day_found(wrong_day: GivenDay) -> Self {
        let day = Some(format!("Solution found for day {}", wrong_day));
        Self {
            day,
            ..Default::default()
        }
    }

    pub fn no_task_found(right_day: GivenDay, wrong_task: GivenTask) -> Self {
        let day = Some(right_day.to_string());
        let task = Some(format!("No solution found for task {}", wrong_task));
        Self {
            day,
            task,
            ..Default::default()
        }
    }

    pub fn average(average: Duration) -> Self {
        Self::new_aggreate(AVERAGE, average)
    }

    pub fn total(total: Duration) -> Self {
        Self::new_aggreate(TOTAL, total)
    }

    pub fn into_row(self) -> Vec<String> {
        fn this_or_placeholder(value: Option<String>) -> String {
            value.unwrap_or_else(|| constants::PLACEHOLDER_IN_BENCHMARK_REPORTS.to_string())
        }

        let (day, task, taken_time, actual_result, expected_result, path) = (
            this_or_placeholder(self.day),
            this_or_placeholder(self.task),
            this_or_placeholder(
                self.taken_time
                    .map(ConductedBenchmark::convert_duration_to_secs_and_mili_txt),
            ),
            this_or_placeholder(self.actual_result),
            this_or_placeholder(self.expected_result),
            this_or_placeholder(self.path_to_input),
        );
        vec![day, task, taken_time, actual_result, expected_result, path]
    }

    fn new_aggreate(label: &str, aggregate: Duration) -> Self {
        let (day, task) = (Some(label.to_string()), Some(label.to_string()));
        let taken_time = Some(aggregate);
        Self {
            day,
            task,
            taken_time,
            ..Default::default()
        }
    }
}
